# Generated by Django 3.0.7 on 2021-03-30 16:59
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0008_profile"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="is_default_version",
            field=models.BooleanField(default=False, verbose_name="Default version"),
        ),
        migrations.AddField(
            model_name="dataset",
            name="version",
            field=models.CharField(
                blank=True, max_length=250, null=True, verbose_name="Schema Version"
            ),
        ),
        migrations.AlterField(
            model_name="datasettable",
            name="db_table",
            field=models.CharField(max_length=100),
        ),
    ]
